;;;; charset.el 0.2c --- charset support

;;; Author: Juliusz Chroboczek <jec@dcs.ed.ac.uk>
;;; Keywords: charset, MIME

;;; This code is copyright (c) 1996 by Juliusz Chroboczek.
;;; Feel free to use and share.

;;; This file provides a first attempt at support for charset handling
;;; in Mail, News, WWW and so on in normal Emacs (not Mule).  It
;;; currently manages to display alphabetic left-to-right text/plain
;;; messages in both TM and RMIME.

;;; Please note that only the functions `charset-dont-munge-p',
;;; `charset-decode-region' and `charset-decode-string' are intended
;;; to be called by User Agents.

(defvar charset-default-charset 'iso-8859-1
  "Default character set used by your Emacs.
Must be a canonical name of a superset of us-ascii.")

(defvar charset-fonts
  '((iso-8859-2 . "-adobe-courier-medium-r-*--14-*-*-*-m-*-iso8859-2")
    (iso-8859-5 . "-cyrillic-courier-medium-r-*-*-10-*-*-*-*-*-iso8859-5"))
  "An association list of charsets and font names.
Canonical charset names only.")

(defvar charset-translations
  '((iso-8859-1 us-ascii L1-ascii . nil)
    (iso-8859-2 us-ascii L2-ascii . nil)
    (windows-1250 iso-8859-2 CP1250-L2 . t)
    (windows-1250 us-ascii (CP1250-L2 L2-ascii) . nil)
    (koi-8 iso-8859-5 koi8-cyrillic . t)
    (iso-8859-5 us-ascii cyrillic-ascii . nil)
    (koi-8 us-ascii (koi8-cyrillic cyrillic-ascii) . nil)
    (ibm850 iso-8859-1 ibm850-L1 . t)
    (ibm850 us-ascii (ibm850-L1 L1-ascii) . nil)
    (iso646-se iso-8859-1 swedish-L1 . t))
  "A list of translations between character encodings.

Elements are of the form
  (FROM TO NAME . WISE-FLAG)
where WISE-FLAG is non-nil when the translation should be preferred.")

(defconst charset-silent
  nil
  "If non-nil charset decoding will be silent")

;;; End of user customizable vars

(defconst charset-digits '("1" "2" "3" "4" "5" "6" "7" "8" "9"))

(defvar charset-aliases
  ;; the first two are really needed
  (cons (cons (intern "") 'us-ascii)
        (append
         '((nil . us-ascii)
           (ascii . us-ascii)
           (us . us-ascii)
           (iso646-us . us-ascii)
           (latin1 . iso-8859-1)
           (l1 . iso-8859-1)
           (latin2 . iso-8859-2)

           (l2 . iso-8859-2)
           (x-cp1250 . windows-1250)
           (x-windows-1250 . windows-1250)
           (x-cp850 . ibm850)
           (cp850 . ibm850)
           (x-ibm850 . ibm850)
           (cyrillic . iso-8859-5)
           (iso646-fi . iso646-se)
           (se . iso646-se)
           (fi . iso646-se))
         (mapcar #'(lambda (x) 
                     (cons (intern (concat "iso_8859-" x))
                           (intern (concat "iso-8859-" x))))
                 charset-digits)))
  "A-list mapping character set aliases to canonical names")

(defvar charset-subsets
  '((us-ascii . iso-8859-1)
    (us-ascii . iso-8859-2)
    (us-ascii . iso-8859-3)
    (us-ascii . iso-8859-4)
    (us-ascii . iso-8859-5)
    (us-ascii . iso-8859-6)
    (us-ascii . iso-8859-7)
    (us-ascii . iso-8859-8)
    (us-ascii . iso-8859-9)
    (us-ascii . koi-8)
    (us-ascii . windows-1250)
    (us-ascii . ibm850))
  "A-list of direct charset subsets; 
must be non looping and containt canonical names only")

(defun charset-make-face (font-name face-name)
  (let ((face-name (intern (concat "charset-default-"
                                   (symbol-name face-name)))))
    (condition-case err
        (progn
          (copy-face 'default face-name)
          (set-face-font face-name font-name)
          face-name)
      (error nil))))


(defvar charset-translation-regexps '()
  "An a-list of cached regexps corresponding to charset translations")

(load "charset-tables")

(defun charset-translation-regexp (table)
  (concat "[" (concat (mapcar #'car table)) "]")) ; XXX

(defun charset-find-translation-regexp (name)
  (or
   (cdr (assoc name charset-translation-regexps))
   (let ((regexp (charset-translation-regexp
                  (charset-find-translation-table name))))
     (if (null regexp)
         nil
       (setq charset-translation-regexps
             (cons (cons name regexp) charset-translation-regexps)))
     regexp)))

(defun charset-find-translation-table (name)
  (let ((full-name (intern (concat "charset-" (symbol-name name)
                                   "-translation-table"))))
    (condition-case err
        (symbol-value full-name)
      (error nil))))

(defun charset-apply-translation (beginning end translation)
  (cond
   ((symbolp translation)
    (charset-apply-translation-1 beginning end translation))
   ((listp translation)
    (while (not (null translation))
      (charset-apply-translation-1 beginning end (car translation))
      (setq translation (cdr translation))))
   (t
    (error "Charset: something impossible just happened"))))

(defun charset-apply-translation-1 (beginning end translation-name)
  (let ((regexp (charset-find-translation-regexp translation-name))
        (table (charset-find-translation-table translation-name)))
    (and regexp table
         (save-excursion
           (goto-char beginning)
           (catch 'charset-translate
             (while t
               (if (re-search-forward regexp end t)
                   (let ((newstring 
                          (cdr (assoc (char-after (match-beginning 0)) 
                                      table))))
                     (if newstring
                         (replace-match newstring t t)))
                 (throw 'charset-translate (point-max)))))))))

(defun charset-canonical (charset)
  "Find the canonical name for charset"
  (or (cdr (assoc charset charset-aliases)) charset))

(defun charset-subset (charset1 charset2)
  "Check whether charset1 is a subset of charset2"
  (charset-subset-canonical (charset-canonical charset1)
                      (charset-canonical charset2)))

(defun charset-subset-canonical (c1 c2)
  (if (eq c1 c2) t
    (catch 'charset-subset
      (let ((remaining-hope charset-subsets))
        (while remaining-hope
          (let ((entry (car remaining-hope)))
            (and (eq (cdr entry) c2)
                 (charset-subset-canonical c1 (car (car remaining-hope)))
                 (throw 'charset-subset t)))
          (setq remaining-hope (cdr remaining-hope)))
        nil))))

(defun charset-face (charset)
  "Find a face for displaying charset directly (canonical name)
Return either a symbol or nil"
  (let ((remaining-hope charset-fonts))
    (catch 'charset-face
      (while remaining-hope
        (let ((entry (car remaining-hope)))
          (if (charset-subset-canonical charset (car entry))
              (if (stringp (cdr entry))
                  (let ((face (charset-make-face (cdr entry) (car entry))))
                    (if (null face)
                        nil
                      (setcdr entry face)
                      (throw 'charset-face (cdr entry))))
                (throw 'charset-face (cdr entry)))))
        (setq remaining-hope (cdr remaining-hope))))))

(defun charset-translation (charset &optional force-wise-flag)
  "Find a translation (no face) for displaying charset (canonical name)"
  (let ((remaining-hope charset-translations))
    (catch 'charset-translation
      (while remaining-hope
        (let ((entry (car remaining-hope)))
          (let ((from (car entry))
                (to (car (cdr entry)))
                (translation (car (cdr (cdr entry))))
                (wise-flag (cdr (cdr (cdr entry)))))
            (and (or wise-flag (not force-wise-flag))
                 (charset-subset-canonical to charset-default-charset)
                 (charset-subset-canonical charset from)
                 (throw 'charset-translation translation))))
        (setq remaining-hope (cdr remaining-hope))))))

(defun charset-translation-face (charset &optional force-wise-flag)
  "Find a translation and a face for displaying charset (canonical name)"
  (let ((remaining-hope charset-translations))
    (catch 'charset-translation-face
      (while remaining-hope
        (let ((entry (car remaining-hope)))
          (let ((from (car entry))
                (to (car (cdr entry)))
                (translation (car (cdr (cdr entry))))
                (wise-flag (cdr (cdr (cdr entry)))))
            (let ((c (and (or wise-flag (not force-wise-flag))
                          (charset-face to))))
              (and c
                   (charset-subset-canonical charset from)
                   (throw 'charset-translation-face
                          (cons translation c))))))
        (setq remaining-hope (cdr remaining-hope))))))

(defun charset-dont-munge-p (charset)
  "Return t if this charset can be displayed as-is"
  (let ((c (charset-canonical charset)))
    (charset-subset-canonical c charset-default-charset)))

(defun charset-decode-method (charset)
  (if (stringp charset)
      (setq charset (intern (downcase charset)))
    (setq charset (intern (downcase (symbol-name charset)))))
  (or
   (if (charset-dont-munge-p charset)
       (cons nil nil))                  ; no problem!
   ;; try with just a font
   (let ((f (charset-face charset)))
     (if f (cons nil f)))
   ;; try with a wise translation
   (let ((tr (charset-translation charset t)))
     (if tr (cons tr nil)))
   (charset-translation-face charset t)
   ;; tough... try an unwise translation
   (let ((tr (charset-translation charset nil)))
     (if tr (cons tr nil)))
   (charset-translation-face charset nil)))
  
(defun charset-decode-region (charset &optional beginning end)
  "Decode region, encoded in charset CHARSET.

Return value is either `nil' if charset could not be
displayed, or (tr . f) where `tr' is the applied translation
and `f' the applied face.  Both might be nil."
  (interactive "SCharset: ")
  (if (stringp charset)
      (setq charset (intern (downcase charset)))
    (setq charset (intern (downcase (symbol-name charset)))))
  (if (and beginning end)
      nil
    (setq beginning (point) end (mark))
    (if (> beginning end)
        (let ((malyjasio beginning))
          (setq beginning end end malyjasio))))
  (setq end (copy-marker end))          ; NOT beginning!
  (let ((method (charset-decode-method charset)))
    (prog1
        (if (null method)
            (if charset-silent
                nil
              (message "Failed to decode charset %s" charset)
              nil)
          (if (car method)
              (charset-apply-translation beginning end (car method)))
          (if (cdr method)
              (put-text-property beginning end 'face (cdr method)))
          method)
      (set-marker end nil))))

(defun charset-decode-string (charset string)
  "Decode STRING, encoded in charset CHARSET"
  (let ((method (charset-decode-method charset)) result)
    (if (null method)
        (if charset-silent
            nil
          (message "Failed to decode charset %s" charset)
          nil)
      (if (null (car method))
          (setq result (copy-sequence string))
        ;; Is it more efficient to use a temporary buffer or to
        ;; mapconcat over the string?
        (save-excursion
          (set-buffer (get-buffer-create "*Charset-Temporary*"))
          (insert string)
          (charset-apply-translation (point-min) (point-max) (car method))
          (setq result (buffer-string))
          (delete-region (point-min) (point-max))))
      (if (cdr method)
          (put-text-property 1 (length result) 'face (cdr method) result))
      result)))

(provide 'charset)
